/*---------------------------------------------------------------------- 
 *      Program dd-HB-dse  
 *----------------------------------------------------------------------
 *
 *  In this test program processor number one reads the whole matrix
 *  from file. The matrix is assumed to be in Harwell-Boeing format.
 *  It  then  partitions  its  graph  using DSE, a  simple partitioning
 *  routine, and scatters the local matrices to each processor. Once
 *  these submatrices are received each processor solves the problem
 *  using preconditioned FGMRES preconditioned with :
 *                       BJ, RAS, SCHUR
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#if defined(__ICC)
#include <mathimf.h>
#else
#include <math.h>
#endif 
#include "parms.h"
#include "aux.h"

int main(int argc, char *argv[])
{
  /* declarations related to Harwell-boeing format for reading the HB 
     matri. Second part is related to I/O parameters */
  char *mname, guesol[2], title[72], key[8], type[3];
  int nrhs, nc, n, nnz, m, tmp0, tmp, tmp2, tmp3, job;
  int myid, ierr, len, i, i1, usym;
  /* memory usage of the preconditioning matrix */
  double ratio;
  /* working array for reading matrix */
  double *a, *rhstmp, wt, norm, res1, tpc, ttol;
  int *ja, *ia, *p;
  int npro, *im, its;
  fprm prm;
  FILE *fp;
  char *name, *iluname;

  /* variables related to dse partitioning */
  int *riord, *dom, *idom, *mask, *jwk, *link;

  /* variables related to pARMS */
  parms_Map       map;
  parms_Vec       x, y, rhs;
  parms_Mat       A;
  parms_PC        pc;
  parms_Solver    solver;
  parms_Timer     tm;

  /* initialize MPI environment */
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  MPI_Comm_size(MPI_COMM_WORLD, &npro);

  tmp0 = 0;
  nrhs = 0;

  /* read matrix name from file input */
  prm = malloc(sizeof(*prm));
  if (prm == NULL) {
    fprintf(stderr, "cannot allocate memory for prm\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }

  if (argc >= 2) {
    read_param(argv[1], prm);
  }
  else if (argc == 1){
    read_param("inputs", prm);
  }


  /* variable "matrix" stores the name of the file in HB format Read a
     Harwell-Boeing matrix. using wreadmtc c-version of sparsit
     routine - call wreadmtc a first time to determine sizes of
     arryas. read in values on the second call.  
  */
  mname = prm->mname;
  len = strlen(mname);
  a = NULL; ja = NULL; ia = NULL; rhstmp = NULL;
  wreadmtc_(&tmp0,&tmp0,&tmp0,mname,&len,a,ja,ia,rhstmp,&nrhs,
	    guesol,&n,&nc,&nnz,title,key,type,&ierr,100,2,72,8,3);     

  a = malloc(nnz*sizeof(*a));
  if (a == NULL) {
    fprintf(stderr, "cannot allocate memory for a\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }
  ja = malloc(nnz*sizeof(*ja));
  if (ja == NULL) {
    fprintf(stderr, "cannot allocate memory for ja\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }
  ia = malloc((n+1)*sizeof(*ia));
  if (ia == NULL) {
    fprintf(stderr, "cannot allocate memory for ia\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }
  rhstmp = malloc(n*sizeof(*rhstmp));
  if (rhstmp == NULL) {
    fprintf(stderr, "cannot allocate memory for rhstmp\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }

  tmp = 2;
  tmp2 = n;
  tmp3 = nnz;
  /* Array sizes determined. Now call wreadmtc again for really
     reading */
  wreadmtc_(&tmp2,&tmp3,&tmp,mname,&len,a,ja,ia,rhstmp,&nrhs,
	    guesol,&n,&nc,&nnz,title,key,type,&ierr,100,2,72,8,3);
  free(rhstmp);
  if(ierr != 0) {
    fprintf(stderr, "ierr = %d\n", ierr);
    fprintf(stderr, "cannot read matrix\n");
    MPI_Finalize();
    exit(1);
  }
  if(myid == 0) {
    if(argc == 3) {
      if (NULL == (fp = fopen(argv[2], "w"))) {
	fprintf(stderr, "cannot open file %s\n", argv[2]);
	exit(1);
      }
    }
    else {
      fp = stdout;
    }

    fprintf(fp, "READ the matrix %.*s %.*s \n",8,key,3,type);
    fprintf(fp, "n = %d, nnz = %d\n\n", n, nnz);
  }
  

  usym = 1;
  if(usym == 1) {
    int *iwk, *jb, *ib, *jc, *ic;
    double *b, *mc;

    iwk = malloc(n*sizeof(*iwk));
    if (iwk == NULL) {
      fprintf(stderr, "cannot allocate memory for iwk\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    b   = malloc(nnz*sizeof(*b));
    if (b == NULL) {
      fprintf(stderr, "cannot allocate memory for b\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    jb  = malloc(nnz*sizeof(*jb));
    if (jb == NULL) {
      fprintf(stderr, "cannot allocate memory for jb\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    ib  = malloc((n+1)*sizeof(*ib));
    if (ib == NULL) {
      fprintf(stderr, "cannot allocate memory for ib\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    mc  = malloc(2*nnz*sizeof(*mc));
    if (mc == NULL) {
      fprintf(stderr, "cannot allocate memory for mc\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    jc  = malloc(2*nnz*sizeof(*jc));
    if (jc == NULL) {
      fprintf(stderr, "cannot allocate memory for jc\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    ic  = malloc((n+1)*sizeof(*ic));
    if (ic == NULL) {
      fprintf(stderr, "cannot allocate memory for ic\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }

    job = 1;
    i = 1;
    csrcsc_(&n, &job, &i, a, ja, ia, b, jb, ib);
    for(i = 0; i < nnz; i++) {
      a[i] = 0.0;
    }
    job = 1; 
    i1 = 2*nnz;
    /* Compute C = A + B */
    aplb_(&n, &n, &job, a, ja, ia, b, jb, ib, mc, jc, ic, &i1, iwk,
	 &ierr);
    nnz = ic[n]-1;
    free(a); free(ja); 

    a = malloc(nnz*sizeof(*a));
    if (a == NULL) {
      fprintf(stderr, "cannot allocate memory for a\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    ja = malloc(nnz*sizeof(*ja));
    if (ja == NULL) {
      fprintf(stderr, "cannot allocate memory for ja\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }

    memcpy(ia, ic, (n+1)*sizeof(*ia));
    memcpy(ja, jc, nnz*sizeof(*ja));
    memcpy(a,  mc, nnz*sizeof(*a));

    free(b);
    free(jb);
    free(ib);
    free(mc);
    free(jc);
    free(ic);
    free(iwk);
  }

  idom = malloc((npro+1)*sizeof(*idom));
  if (idom == NULL) {
    fprintf(stderr, "cannot allocate memory for idom\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }
  dom = malloc(n*sizeof(*dom));
  if (dom == NULL) {
    fprintf(stderr, "cannot allocate memory for dom\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }

  if (npro == 1) {
    for (i = 0; i < n; i++) {
      dom[i] = i+1;
    }
    idom[0] = 1;
    idom[1] = n+1;
  }
  else {
    riord = malloc(n*sizeof(*riord));
    if (riord == NULL) {
      fprintf(stderr, "cannot allocate memory for riord\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    mask = malloc(n*sizeof(*mask));
    if (mask == NULL) {
      fprintf(stderr, "cannot allocate memory for mask\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    jwk = malloc(2*n*sizeof(*jwk));
    if (jwk == NULL) {
      fprintf(stderr, "cannot allocate memory for jwk\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    link = malloc(n*sizeof(*link));
    if (link == NULL) {
      fprintf(stderr, "cannot allocate memory for link\n");
      MPI_Abort(MPI_COMM_WORLD, 66);
    }
    dse_(&n, ja, ia, &npro, riord, dom, idom, mask, jwk, link);
    free(riord);
    free(mask);
    free(jwk);
    free(link);
  }

  /* Create map object */
  parms_MapCreateFromPtr(&map, n, dom, idom, MPI_COMM_WORLD, 1, NONINTERLACED);
  
  /* Free dom and idom */
  free(dom);
  free(idom);

  /* Create distributed vectors based on map */
  parms_VecCreate(&x,   map);
  parms_VecCreate(&y,   map);
  parms_VecCreate(&rhs, map);
  parms_VecPut(x, 1.0);

  /* create a distributed matrix based on map */
  parms_MatCreate(&A, map);

  /* Insert values into A */
  PARMS_NEWARRAY(im, n);
  im = malloc(n*sizeof(*im));
  if (im == NULL) {
    fprintf(stderr, "cannot allocate memory for im\n");
    MPI_Abort(MPI_COMM_WORLD, 66);
  }
  for (i = 0; i < n; i++) {
    im[i] = i+1;
  }
  parms_MatSetValues(A, n, im, ia, ja, a, INSERT);

  /* free the matrix stored in CSR format(a,ja,ia) */
  free(a);
  free(ja);
  free(ia);
  free(im);

  /* Setup the matrix and communication structure */
  parms_MatSetup(A);

  /* Setup artifical right-hand-side vector */
  parms_MatApply(A, x, rhs);

  /* Setup initial guess to a vector of all-0. */
  parms_VecPut(x, 0.0);

  /* Get 2-norm of initial residual */
  parms_VecGetNorm2(rhs, &norm);

  /* Create preconditioner based on the matrix A. */
  parms_PCCreate(&pc, A);

  /* set parameters for pc */
  set_pc_params(pc, prm);

  /* create a timer */
  parms_TimerCreate(&tm);
  /* reset the timer */
  parms_TimerReset(tm);
  parms_PCSetup(pc);

  /* get the elapsed time spent on creating PC */
  tpc =  parms_TimerGet(tm);

  /* get the ratio of the number of nonzero entries in the
     preconditioning matrix to that in the original matrix */
  parms_PCGetRatio(pc, &ratio);

  /* pause the timer */
  parms_TimerPause(tm);

  /* Create a solver based on A and pc */
  parms_SolverCreate(&solver, A, pc);

  /* Set the solver type */
  parms_SolverSetType(solver, SOLFGMRES);

  /* set parameters for solver */
  set_solver_params(solver, prm);

  free(prm);

  /* set up solver */
  parms_SolverSetup(solver);

  /* restart the timer */
  parms_TimerRestart(tm);

  /* Solve the linear equation */
  parms_SolverApply(solver, rhs, x);

  /* get total time spent on creating the pc and solving the linear
     system */
  ttol = parms_TimerGet(tm);

  /* Get the number of iterations */
  its = parms_SolverGetIts(solver);

  /* Compute the residual error  */
  parms_MatMVPY(A, -1.0, x, 1.0, rhs, y);
  parms_VecGetNorm2(y, &res1);

  /* processor 0 outputs the result */
  if (myid == 0) {
    char buf[40];
    parms_PCGetName(pc, &name);
    parms_PCILUGetName(pc, &iluname);
    fprintf(fp, "The preconditioner  %9s %s\n", " ", name);
    fprintf(fp, "The local preconditioner %4s %s\n", " ", iluname);
    fprintf(fp, "The memory usage %12s %-4.2f\n", "=", ratio);
    fprintf(fp, "The number of processors %4s %-4d\n", "=", npro);
    fprintf(fp, "The number of iterations %4s %-4d\n", "=", its);
    sprintf(buf, "%8.2fs", tpc);
    fprintf(fp, "The time for pc setup %7s %-s\n", "=", strtok(buf, " "));
    sprintf(buf, "%8.2fs", ttol-tpc);
    fprintf(fp, "The solving time %12s %-s\n", "=", strtok(buf, " "));
    sprintf(buf, "%8.2fs", ttol);
    fprintf(fp, "The total time %14s %-s\n", "=", strtok(buf, " "));
    fprintf(fp, "The initial residual error  = %-8.2e\n", norm);
    fprintf(fp, "The solution residual error = %-8.2e\n", res1);

    if(fp != stdout) {
      fclose(fp);
    }
  }
  /* Free memories */
  parms_VecFree(&x);
  parms_VecFree(&y);
  parms_VecFree(&rhs);
  parms_MatFree(&A);
  parms_MapFree(&map);
  parms_PCFree(&pc);
  parms_SolverFree(&solver);
  parms_TimerFree(&tm);

  /* Exit MPI environment */
  MPI_Finalize();  
}
